/*
 * Copyright © 2013 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *     Antti Kaijanmäki <antti.kaijanmaki@canonical.com>
 */

#include "link.h"
#include "access-point.h"

using namespace platform::nmofono::wifi;
namespace networking = com::ubuntu::connectivity::networking;
namespace dbus = core::dbus;
namespace fdo = org::freedesktop;
namespace NM = fdo::NetworkManager;

using core::Property;

Link::Link(NM::Interface::Device dev,
           NM::Interface::NetworkManager nm)
    : m_dev(dev),
      m_wireless(dev.object),
      m_nm(nm)
{   
    m_charasteristics.set(0);
    m_status.set(Status::disabled);

    std::set<std::shared_ptr<networking::wifi::AccessPoint>> list;
    for (auto access_point : dev.get_access_points()) {
        list.insert(std::make_shared<AccessPoint>(access_point));
    }
    m_accessPoints.set(list);

    m_dev.state_changed->connect([](NM::Interface::Device::Signal::StateChanged::ArgumentType args) {
        // StateChanged ( u: new_state, u: old_state, u: reason )
#if 0
        typedef enum {
            NM_DEVICE_STATE_UNKNOWN      = 0,
            NM_DEVICE_STATE_UNMANAGED    = 10,
            NM_DEVICE_STATE_UNAVAILABLE  = 20,
            NM_DEVICE_STATE_DISCONNECTED = 30,
            NM_DEVICE_STATE_PREPARE      = 40,
            NM_DEVICE_STATE_CONFIG       = 50,
            NM_DEVICE_STATE_NEED_AUTH    = 60,
            NM_DEVICE_STATE_IP_CONFIG    = 70,
            NM_DEVICE_STATE_IP_CHECK     = 80,
            NM_DEVICE_STATE_SECONDARIES  = 90,
            NM_DEVICE_STATE_ACTIVATED    = 100,
            NM_DEVICE_STATE_DEACTIVATING = 110,
            NM_DEVICE_STATE_FAILED       = 120
        } NMDeviceState;

        /*
         * Device state change reason codes
         */
        typedef enum {
            /* No reason given */
            NM_DEVICE_STATE_REASON_NONE = 0,

            /* Unknown error */
            NM_DEVICE_STATE_REASON_UNKNOWN = 1,

            /* Device is now managed */
            NM_DEVICE_STATE_REASON_NOW_MANAGED = 2,

            /* Device is now unmanaged */
            NM_DEVICE_STATE_REASON_NOW_UNMANAGED = 3,

            /* The device could not be readied for configuration */
            NM_DEVICE_STATE_REASON_CONFIG_FAILED = 4,

            /* IP configuration could not be reserved (no available address, timeout, etc) */
            NM_DEVICE_STATE_REASON_IP_CONFIG_UNAVAILABLE = 5,

            /* The IP config is no longer valid */
            NM_DEVICE_STATE_REASON_IP_CONFIG_EXPIRED = 6,

            /* Secrets were required, but not provided */
            NM_DEVICE_STATE_REASON_NO_SECRETS = 7,

            /* 802.1x supplicant disconnected */
            NM_DEVICE_STATE_REASON_SUPPLICANT_DISCONNECT = 8,

            /* 802.1x supplicant configuration failed */
            NM_DEVICE_STATE_REASON_SUPPLICANT_CONFIG_FAILED = 9,

            /* 802.1x supplicant failed */
            NM_DEVICE_STATE_REASON_SUPPLICANT_FAILED = 10,

            /* 802.1x supplicant took too long to authenticate */
            NM_DEVICE_STATE_REASON_SUPPLICANT_TIMEOUT = 11,

            /* PPP service failed to start */
            NM_DEVICE_STATE_REASON_PPP_START_FAILED = 12,

            /* PPP service disconnected */
            NM_DEVICE_STATE_REASON_PPP_DISCONNECT = 13,

            /* PPP failed */
            NM_DEVICE_STATE_REASON_PPP_FAILED = 14,

            /* DHCP client failed to start */
            NM_DEVICE_STATE_REASON_DHCP_START_FAILED = 15,

            /* DHCP client error */
            NM_DEVICE_STATE_REASON_DHCP_ERROR = 16,

            /* DHCP client failed */
            NM_DEVICE_STATE_REASON_DHCP_FAILED = 17,

            /* Shared connection service failed to start */
            NM_DEVICE_STATE_REASON_SHARED_START_FAILED = 18,

            /* Shared connection service failed */
            NM_DEVICE_STATE_REASON_SHARED_FAILED = 19,

            /* AutoIP service failed to start */
            NM_DEVICE_STATE_REASON_AUTOIP_START_FAILED = 20,

            /* AutoIP service error */
            NM_DEVICE_STATE_REASON_AUTOIP_ERROR = 21,

            /* AutoIP service failed */
            NM_DEVICE_STATE_REASON_AUTOIP_FAILED = 22,

            /* The line is busy */
            NM_DEVICE_STATE_REASON_MODEM_BUSY = 23,

            /* No dial tone */
            NM_DEVICE_STATE_REASON_MODEM_NO_DIAL_TONE = 24,

            /* No carrier could be established */
            NM_DEVICE_STATE_REASON_MODEM_NO_CARRIER = 25,

            /* The dialing request timed out */
            NM_DEVICE_STATE_REASON_MODEM_DIAL_TIMEOUT = 26,

            /* The dialing attempt failed */
            NM_DEVICE_STATE_REASON_MODEM_DIAL_FAILED = 27,

            /* Modem initialization failed */
            NM_DEVICE_STATE_REASON_MODEM_INIT_FAILED = 28,

            /* Failed to select the specified APN */
            NM_DEVICE_STATE_REASON_GSM_APN_FAILED = 29,

            /* Not searching for networks */
            NM_DEVICE_STATE_REASON_GSM_REGISTRATION_NOT_SEARCHING = 30,

            /* Network registration denied */
            NM_DEVICE_STATE_REASON_GSM_REGISTRATION_DENIED = 31,

            /* Network registration timed out */
            NM_DEVICE_STATE_REASON_GSM_REGISTRATION_TIMEOUT = 32,

            /* Failed to register with the requested network */
            NM_DEVICE_STATE_REASON_GSM_REGISTRATION_FAILED = 33,

            /* PIN check failed */
            NM_DEVICE_STATE_REASON_GSM_PIN_CHECK_FAILED = 34,

            /* Necessary firmware for the device may be missing */
            NM_DEVICE_STATE_REASON_FIRMWARE_MISSING = 35,

            /* The device was removed */
            NM_DEVICE_STATE_REASON_REMOVED = 36,

            /* NetworkManager went to sleep */
            NM_DEVICE_STATE_REASON_SLEEPING = 37,

            /* The device's active connection disappeared */
            NM_DEVICE_STATE_REASON_CONNECTION_REMOVED = 38,

            /* Device disconnected by user or client */
            NM_DEVICE_STATE_REASON_USER_REQUESTED = 39,

            /* Carrier/link changed */
            NM_DEVICE_STATE_REASON_CARRIER = 40,

            /* The device's existing connection was assumed */
            NM_DEVICE_STATE_REASON_CONNECTION_ASSUMED = 41,

            /* The supplicant is now available */
            NM_DEVICE_STATE_REASON_SUPPLICANT_AVAILABLE = 42,

            /* The modem could not be found */
            NM_DEVICE_STATE_REASON_MODEM_NOT_FOUND = 43,

            /* The Bluetooth connection failed or timed out */
            NM_DEVICE_STATE_REASON_BT_FAILED = 44,

            /* GSM Modem's SIM Card not inserted */
            NM_DEVICE_STATE_REASON_GSM_SIM_NOT_INSERTED = 45,

            /* GSM Modem's SIM Pin required */
            NM_DEVICE_STATE_REASON_GSM_SIM_PIN_REQUIRED = 46,

            /* GSM Modem's SIM Puk required */
            NM_DEVICE_STATE_REASON_GSM_SIM_PUK_REQUIRED = 47,

            /* GSM Modem's SIM wrong */
            NM_DEVICE_STATE_REASON_GSM_SIM_WRONG = 48,

            /* InfiniBand device does not support connected mode */
            NM_DEVICE_STATE_REASON_INFINIBAND_MODE = 49,

            /* A dependency of the connection failed */
            NM_DEVICE_STATE_REASON_DEPENDENCY_FAILED = 50,

            /* Problem with the RFC 2684 Ethernet over ADSL bridge */
            NM_DEVICE_STATE_REASON_BR2684_FAILED = 51,

            /* ModemManager not running */
            NM_DEVICE_STATE_REASON_MODEM_MANAGER_UNAVAILABLE = 52,

            /* The WiFi network could not be found */
            NM_DEVICE_STATE_REASON_SSID_NOT_FOUND = 53,

            /* A secondary connection of the base connection failed */
            NM_DEVICE_STATE_REASON_SECONDARY_CONNECTION_FAILED = 54,

            /* Unused */
            NM_DEVICE_STATE_REASON_LAST = 0xFFFF
        } NMDeviceStateReason;
#endif
        std::cout << "Device State Changed: new: " << std::get<0>(args) << " old: " << std::get<1>(args) << " reason: " << std::get<2>(args) << std::endl;
    });

    m_wireless.access_point_added->connect([this](const dbus::types::ObjectPath &path) {
        for (auto ap : m_accessPoints.get()) {
            if (std::dynamic_pointer_cast<AccessPoint>(ap)->m_ap.object->path() == path) {
                // already in the list
                return;
            }
        }

        auto list = m_accessPoints.get();
        NM::Interface::AccessPoint ap(m_nm.service->object_for_path(path));
        list.insert(std::make_shared<AccessPoint>(ap));
        m_accessPoints.set(list);
    });
    m_wireless.access_point_removed->connect([this](const dbus::types::ObjectPath &path) {
        auto list = m_accessPoints.get();
        for (auto ap : list) {
            if (std::dynamic_pointer_cast<AccessPoint>(ap)->m_ap.object->path() == path) {
                list.erase(ap);
                m_accessPoints.set(list);
                return;
            }
        }
    });
}

Link::~Link()
{}


void
Link::enable()
{
    std::cout << "Link::enable()" << std::endl;

    /** @todo:
     * NetworkManager 0.9.9 allows doing this below and letting NM to figure out the best connection:
     *
     * m_nm.activate_connection(dbus::types::ObjectPath("/"),
     *                          m_dev.object->path(),
     *                          core::dbus::typ es::ObjectPath("/"));
     *
     * for now we get somewhat similar result by toggling the Autoconnect property,
     * but this breaks the possible setting we will have for autoconnecting later.
     */
    m_dev.autoconnect->set(true);

}

void
Link::disable()
{
    std::cout << "Link::disable()" << std::endl;

    /** @todo remove this after NM 0.9.9 is available. check comment in enable() */
    m_dev.autoconnect->set(false);

    m_dev.disconnect();
    m_status.set(Status::disabled);
}

networking::Link::Type
Link::type() const
{
    return Type::wifi;
}

const Property<int>&
Link::charasteristics()
{
    return m_charasteristics;
}

const Property<networking::Link::Status>&
Link::status()
{
    return m_status;
}

networking::Link::Id
Link::id() const
{
    return 0;
}

std::string
Link::name() const
{
    return "";
}

const Property<std::set<std::shared_ptr<networking::wifi::AccessPoint> > >&
Link::accessPoints()
{
    return m_accessPoints;
}

void
Link::connect_to(std::shared_ptr<networking::wifi::AccessPoint> accessPoint)
{
    std::shared_ptr<AccessPoint> ap = std::dynamic_pointer_cast<AccessPoint>(accessPoint);
    std::cout << "Connect to: " << accessPoint->ssid() << std::endl;

    NM::Interface::Connection *found = nullptr;
    auto connections = m_dev.get_available_connections();
    for (auto &con : connections) {
        for (auto map : con.get_settings()) {
            if (map.first == "802-11-wireless") {
                for (auto conf : map.second) {
                    if (conf.first == "ssid") {
                        std::vector<int8_t> value;
                        value = conf.second.as<std::vector<std::int8_t>>();
                        if (value == ap->m_ap.ssid->get()) {
                            found = &con;
                            break;
                        }
                    }
                }
            }
        }
    }

    /// @todo check the timestamps as there might be multiple ones that are suitable.
    /// @todo oh, and check more parameters than just the ssid

    if (found) {
        m_nm.activate_connection(found->object->path(),
                                 m_dev.object->path(),
                                 ap->m_ap.object->path());
    } else {
        std::map<std::string, std::map<std::string, dbus::types::Variant> > conf;

        /// @todo getting the ssid multiple times over dbus is stupid.

        std::map<std::string, dbus::types::Variant> wireless_conf;
        wireless_conf["ssid"] = dbus::types::Variant::encode<std::vector<std::int8_t>>(ap->m_ap.ssid->get());

        conf["802-11-wireless"] = wireless_conf;
        m_nm.add_and_activate_connection(conf,
                                         m_dev.object->path(),
                                         ap->m_ap.object->path());
    }
}

const Property<std::shared_ptr<networking::wifi::AccessPoint> >&
Link::activeAccessPoint()
{
    return m_activeAccessPoint;
}

void
Link::request_scan()
{

}
